/* VzImagination (C) 2018-2019 Stephane Charette <stephanecharette@gmail.com>
 * $Id: TestMatManip.cpp 2741 2019-04-09 17:04:20Z stephane $
 */

#include "TestUtil.hpp"


TEST(Invert, Invert)
{
	// Inverting an image twice should get us back to the original image.

	for (const auto & [path, name] : complex_test_images())
	{
		std::cout << name << std::endl;

		cv::Mat img1 = cv::imread(path);
		cv::Mat img2 = vz::invert(img1);
		cv::Mat img3 = vz::invert(img2);

		ASSERT_TRUE(vz::compare_exact(img1, img3));
	}
}


TEST(CannyEdgeDetection, CannyEdgeDetection)
{
	for (const auto & [path, name] : complex_test_images())
	{
		vz::Imagination img(path);

		cv::Mat threshold = img.get(vz::Imagination::EMatType::kThreshold);
		double d = img.threshold_value;

		cv::Mat canny1 = vz::apply_canny_edge_detection(threshold, d);
		cv::Mat canny2 = vz::apply_canny_edge_detection(img.get(vz::Imagination::EMatType::kOriginalImage		), d);
		cv::Mat canny3 = vz::apply_canny_edge_detection(img.get(vz::Imagination::EMatType::kO2GaussianBlur, 3	), d);
		cv::Mat canny4 = vz::apply_canny_edge_detection(img.get(vz::Imagination::EMatType::kO2GaussianBlur, 15	), d);

		cv::imwrite("canny_edge_" + name + "_using_threshold.jpg"	, canny1, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
		cv::imwrite("canny_edge_" + name + "_using_original.jpg"	, canny2, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
		cv::imwrite("canny_edge_" + name + "_using_blur_03.jpg"		, canny3, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
		cv::imwrite("canny_edge_" + name + "_using_blur_15.jpg"		, canny4, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
	}
}


TEST(Threshold, AllThresholds)
{
	for (const auto & [path, name] : complex_test_images())
	{
		std::cout << "finding threshold for " << name << "..." << std::endl;

		cv::Mat src = cv::imread(path);
		cv::imwrite("threshold_" + name + "_t0_1_original.jpg", src, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});

		cv::Mat colour_balanced = vz::apply_simple_colour_balance(src);
//		cv::imwrite("threshold_" + name + "_t0_2_balanced.jpg", colour_balanced, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});

		cv::Mat blur = vz::apply_gaussian_blur(colour_balanced);
//		cv::imwrite("threshold_" + name + "_t0_3_blur.jpg", blur, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});

		cv::Mat grey;
		cv::cvtColor(blur, grey, cv::COLOR_BGR2GRAY);
//		cv::imwrite("threshold_" + name + "_t0_4_grey.jpg", grey, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});

		const auto total = grey.cols * grey.rows;
		vz::VContours contours;

		cv::Mat t1;
		cv::Mat c1;
		cv::Mat z1;
		const double d1 = cv::threshold(grey, t1, 64, 255, cv::THRESH_BINARY | cv::THRESH_OTSU);
		auto white = cv::countNonZero(t1);
		cv::Canny(t1, c1, d1, 3.0 * d1, 3, true);
		contours = vz::find_contours(t1);
		z1 = vz::add_contours_to_image(src, contours);
		cv::imwrite("threshold_" + name + "_t1_otsu.png"			, t1, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t1_otsu_canny.png"		, c1, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t1_otsu_contours.png"	, z1, {cv::IMWRITE_PNG_COMPRESSION, 9});
		std::cout << name << ": d1=" << d1 << std::endl;
		std::cout << name << ": T1 count: " << white << " white out of " << total << " " << ((double)white / (double)total * 100.0) << "%" << std::endl;
		std::cout << name << ": T1 contours: " << contours.size() << std::endl;

		cv::Mat t2;
		cv::Mat c2;
		cv::Mat z2;
		const double d2 = cv::threshold(grey, t2, 64, 255, cv::THRESH_BINARY | cv::THRESH_TRIANGLE);
		white = cv::countNonZero(t2);
		cv::Canny(t2, c2, d2, 3.0 * d2, 3, true);
		contours = vz::find_contours(t2);
		z2 = vz::add_contours_to_image(src, contours);
		cv::imwrite("threshold_" + name + "_t2_triangle.png"			, t2, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t2_triangle_canny.png"		, c2, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t2_triangle_contours.png"	, z2, {cv::IMWRITE_PNG_COMPRESSION, 9});
		std::cout << name << ": d2=" << d2 << std::endl;
		std::cout << name << ": T2 count: " << white << " white out of " << total << " " << ((double)white / (double)total * 100.0) << "%" << std::endl;
		std::cout << name << ": T2 contours: " << contours.size() << std::endl;

		cv::Mat t3;
		cv::Mat c3;
		cv::Mat z3;
		cv::adaptiveThreshold(grey, t3, 255, cv::ADAPTIVE_THRESH_GAUSSIAN_C, cv::THRESH_BINARY, 5, 9.0);
		white = cv::countNonZero(t3);
		cv::Canny(t3, c3, d1, 3.0 * d1, 3, true);
		contours = vz::find_contours(t3);
		z3 = vz::add_contours_to_image(src, contours);
		cv::imwrite("threshold_" + name + "_t3_gaussian.png"			, t3, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t3_gaussian_canny.png"		, c3, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t3_gaussian_contours.png"	, z3, {cv::IMWRITE_PNG_COMPRESSION, 9});
		std::cout << name << ": T3 count: " << white << " white out of " << total << " " << ((double)white / (double)total * 100.0) << "%" << std::endl;
		std::cout << name << ": T3 contours: " << contours.size() << std::endl;

		cv::Mat t4;
		cv::Mat c4;
		cv::Mat z4;
		cv::adaptiveThreshold(grey, t4, 255, cv::ADAPTIVE_THRESH_MEAN_C, cv::THRESH_BINARY, 5, 9.0);
		white = cv::countNonZero(t4);
		cv::Canny(t4, c4, d1, 3.0 * d1, 3, true);
		contours = vz::find_contours(t4);
		z4 = vz::add_contours_to_image(src, contours);
		cv::imwrite("threshold_" + name + "_t4_mean.png"			, t4, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t4_mean_canny.png"		, c4, {cv::IMWRITE_PNG_COMPRESSION, 9});
		cv::imwrite("threshold_" + name + "_t4_mean_contours.png"	, z4, {cv::IMWRITE_PNG_COMPRESSION, 9});
		std::cout << name << ": T4 count: " << white << " white out of " << total << " " << ((double)white / (double)total * 100.0) << "%" << std::endl;
		std::cout << name << ": T4 contours: " << contours.size() << std::endl;

		vz::Imagination img(path);
		cv::Mat t5 = img.get(vz::Imagination::EMatType::kThreshold);
		white = cv::countNonZero(t5);
		contours = vz::find_contours(t5);
		cv::imwrite("threshold_" + name + "_t5_imagination.png", t5, {cv::IMWRITE_PNG_COMPRESSION, 9});
		std::cout << name << ": T5 count: " << white << " white out of " << total << " " << ((double)white / (double)total * 100.0) << "%" << std::endl;
		std::cout << name << ": T5 contours: " << contours.size() << std::endl;
	}
}


TEST(Histogram, Histogram)
{
	for (const auto & [path, name] : test_images())
	{
		cv::Mat src = cv::imread(path);
		cv::Mat greyscale = vz::to_greyscale(src);

		cv::Mat histogram1 = vz::plot_histogram(src, 320, 240);
		ASSERT_EQ(histogram1.channels(), 3);
		ASSERT_EQ(histogram1.cols, 320);
		ASSERT_EQ(histogram1.rows, 240);

		cv::Mat histogram2 = vz::plot_histogram(greyscale, 320, 240);
		ASSERT_EQ(histogram2.channels(), 3);
		ASSERT_EQ(histogram2.cols, 320);
		ASSERT_EQ(histogram2.rows, 240);

		cv::Mat histogram3 = vz::plot_histogram(src);
		ASSERT_EQ(histogram3.channels(), 3);
		ASSERT_EQ(histogram3.cols, 640);
		ASSERT_EQ(histogram3.rows, 480);

		ASSERT_FALSE(vz::compare_exact(histogram1, histogram2));
		ASSERT_FALSE(vz::compare_exact(histogram1, histogram3));
		ASSERT_FALSE(vz::compare_exact(histogram2, histogram3));

		cv::imwrite("histogram_" + name + ".png", histogram3, {cv::IMWRITE_PNG_COMPRESSION, 9});
	}
}


TEST(Histogram, Invalid)
{
	test_images();

	// cannot create a histogram with an empty image
	cv::Mat mat;
	ASSERT_THROW(vz::plot_histogram(mat), std::exception);

	// cannot create a histogram with a too-small image size
	mat = cv::imread("capture.jpg");
	ASSERT_THROW(vz::plot_histogram(mat, 9, 9), std::exception);
}
